from .basic_api_errors import BasicAPIErrorTest
from .basic_api_test import BasicAPITest
from .test_auth_api import AuthAPITest
from .test_cache import TestCache
from .test_fields import FieldsTest
from .test_locale_api import LocaleAPITest
from .test_perm_tests import PermTestsTest
from .test_perms_api import PermsAPITest
from .test_shortcuts import ShortcutTest
from .test_user_queryset import TestUserQuerySet
from .test_view_fields import ViewFieldTest
